"""Data structures for the status update query communication."""

import datetime
from enum import Enum
import typing

from pydantic import BaseModel
from pydantic.config import ConfigDict


class ErrorCode(Enum):
    SUCCESS = "0"
    PARTIAL_SUCCESS = "1"
    ALL_FAILURE = "2"


class ProcessingStatus(Enum):
    ERROR = "Error"
    SUCCESS = "Success"


class SubmissionStatusFile(BaseModel):
    """Type for ``SubmissionStatus`` entry ``actions[*].response[*].files[*]``."""

    model_config = ConfigDict(frozen=True)

    #: File URL
    url: str


class SubmissionStatusObjectContent(BaseModel):
    """type for ``SubmissionStatusObjectContent`` entry in ``actions[*].response[*].objects[*].content``."""

    model_config = ConfigDict(frozen=True)

    #: Processing status
    clinvarProcessingStatus: str
    #: Release status
    clinvarReleaseStatus: str


class SubmissionStatusObject(BaseModel):
    """Type for ``SubmissionStatusObject`` entry in ``actions[*].response[*].objects[*]``."""

    model_config = ConfigDict(frozen=True)

    #: Optional object accession.
    accession: typing.Optional[str]
    #: Object content.
    content: SubmissionStatusObjectContent
    #: Target database, usually "clinvar" per the docs.
    targetDb: str


class SubmissionStatusResponseMessage(BaseModel):
    """Type for ``SubmissionStatusResponseMessage`` entry in ``actions[*].response[*].message``."""

    model_config = ConfigDict(frozen=True)

    #: The error code.
    errorCode: typing.Optional[ErrorCode]
    #: The message severity.
    severity: str
    #: The message text.
    text: str


class SubmissionStatusResponse(BaseModel):
    """Type for ``SubmissionStatus`` entry ``actions[*].response[*]``."""

    model_config = ConfigDict(frozen=True)

    #: Status, one of "processing", "processed", "error",
    status: str
    #: Files
    files: typing.List[SubmissionStatusFile]
    #: Message
    message: typing.Optional[SubmissionStatusResponseMessage]
    #: Objects
    objects: typing.List[SubmissionStatusObject]


class SubmissionStatusActions(BaseModel):
    """Type for ``SubmissionStatus`` entry ``actions[*]``."""

    model_config = ConfigDict(frozen=True)

    #: Identifier of the submission
    id: str
    #: Entries in ``actions[*].responses``, only one entry per the docs.
    responses: typing.List[SubmissionStatusResponse]
    #: Status of the submission, one of "submitted", "processing", "processed", "error"
    status: str
    #: Target database, usually "clinvar"
    targetDb: str
    #: Last updated time
    updated: datetime.datetime


class SubmissionStatus(BaseModel):
    """Representation of server's response to a submission status query."""

    model_config = ConfigDict(frozen=True)

    #: The list of actions (one element only by the docs).
    actions: typing.List[SubmissionStatusActions]



"""Data structures for encoding the submission payload."""

from enum import Enum
import typing

from pydantic import BaseModel
from pydantic.config import ConfigDict


class Assembly(Enum):
    GRCH38 = "GRCh38"
    HG38 = "hg38"
    GRCH37 = "GRCh37"
    HG19 = "hg19"
    NCBI36 = "NCBI36"
    HG18 = "hg18"


class Chromosome(Enum):
    CHR1 = "1"
    CHR2 = "2"
    CHR3 = "3"
    CHR4 = "4"
    CHR5 = "5"
    CHR6 = "6"
    CHR7 = "7"
    CHR8 = "8"
    CHR9 = "9"
    CHR10 = "10"
    CHR11 = "11"
    CHR12 = "12"
    CHR13 = "13"
    CHR14 = "14"
    CHR15 = "15"
    CHR16 = "16"
    CHR17 = "17"
    CHR18 = "18"
    CHR19 = "19"
    CHR20 = "20"
    CHR21 = "21"
    CHR22 = "22"
    CHRX = "X"
    CHRY = "Y"
    CHRMT = "MT"


class VariantType(Enum):
    INSERTION = "Insertion"
    DELETION = "Deletion"
    DUPLICATION = "Duplication"
    TANDEM_DUPLICATIOn = "Tandem duplication"
    COPY_NUMBER_LOSS = "copy number loss"
    COPY_NUMBER_GAIN = "copy number gain"
    INVERSION = "Inversion"
    TRANSLOCATION = "Translocation"
    COMPLEX = "Complex"


class CitationDb(Enum):
    """Allowed values for a citation's `db` field.

    The values of the enumeration map to the values used by the ClinVar submission API.
    """

    PUBMED = "PubMed"
    BOOKSHELF = "BookShelf"
    DOI = "DOI"
    PMC = "pmc"


class ConditionDb(Enum):
    OMIM = "OMIM"
    MEDGEN = "MedGen"
    ORPHANET = "Orphanet"
    MESH = "MeSH"
    HP = "HP"
    MONDO = "MONDO"


class MultipleConditionExplanation(Enum):
    NOVEL_DISEASE = "Novel disease"
    UNCERTAIN = "Uncertain"
    CO_OCCURING = "Co-occurring"


class AffectedStatus(Enum):
    YES = "yes"
    NO = "no"
    UNKNOWN = "unknown"
    NOT_PROVIDED = "not provided"
    NOT_APPLICABLE = "not applicable"


class AlleleOrigin(Enum):
    GERMLINE = "germline"
    SOMATIC = "somatic"
    DE_NOVO = "de novo"
    UNKNOWN = "unknown"
    INHERITED = "inherited"
    MATERNAL = "maternal"
    PATERNAL = "paternal"
    BIPARENTL = "biparental"
    NOT_APPLICABLE = "not applicable"


class ClinicalFeaturesAffectedStatus(Enum):
    PRESENT = "present"
    ABSENT = "absent"
    NOT_TESTED = "not tested"


class ClinicalFeaturesDb(Enum):
    HP = "HP"


class CollectionMethod(Enum):
    CURATION = "curation"
    LITERATURE_ONLY = "literature only"
    REFERENCE_POPULATION = "reference population"
    PROVIDER_INTERPRETATION = "provider interpretation"
    PHENOTYPING_ONLY = "phenotyping only"
    CASE_CONTROL = "case-control"
    CLINICAL_TESTING = "clinical testing"
    IN_VITRO = "in vitro"
    IN_VIVO = "in vivo"
    RESEARCH = "research"
    NOT_PROVIDED = "not provided"


class StructVarMethodType(Enum):
    SNP_ARRAY = "SNP array"
    OLIGO_ARRAY = "Oligo array"
    READ_DEPTH = "Read depth"
    PAIRED_END_MAPPING = "Paired-end mapping"
    ONE_END_ANCHORED_ASSEMBLY = "One end anchored assembly"
    SEQUENCE_ALIGNMENT = "Sequence alignment"
    OPTICAL_MAPPING = "Optical mapping"
    CURATED_PCR = "Curated,PCR"


class ClinicalSignificanceDescription(Enum):
    """Allowed values for the ``clinicalSignificanceDescription``.

    The values of the enumeration map to the values used by the ClinVar submission API.
    """

    PATHOGENIC = "Pathogenic"
    LIKELY_PATHOGENIC = "Likely pathogenic"
    UNCERTAIN_SIGNIFICANCE = "Uncertain significance"
    LIKELY_BENIGN = "Likely benign"
    BENIGN = "Benign"
    PATHOGENIC_LOW_PENETRANCE = "Pathogenic, low penetrance"
    UNCERTAIN_RISK_ALLELE = "Uncertain risk allele"
    LIKELY_PATHOGENIC_LOW_PENETRANCE = "Likely pathogenic, low penetrance"
    ESTABLISHED_RISK_ALLELE = "Established risk allele"
    LIKELY_RISK_ALLELE = "Likely risk allele"
    AFFECTED = "affects"
    ASSOCIATION = "association"
    DRUG_RESPONSE = "drug response"
    CONFERS_SENSITIVITY = "confers sensitivity"
    PROTECTIVE = "protective"
    OTHER = "other"
    NOT_PROVIDED = "not provided"


class ModeOfInheritance(Enum):
    AUTOSOMAL_DOMINANT_INHERITANCE = "Autosomal dominant inheritance"
    AUTOSOMAL_RECESSIVE_INHERITANCE = "Autosomal recessive inheritance"
    MITOCHONDRIAL_INHERITANCE = "Mitochondrial inheritance"
    SOMATIC_MUTATION = "Somatic mutation"
    GENETIC_ANTICIPATION = "Genetic anticipation"
    SPORADIC = "Sporadic"
    SEX_LIMITED_AUTOSOMAL_DOMINANT = "Sex-limited autosomal dominant"
    X_LINKED_RECESSIVE_INHERITANCE = "X-linked recessive inheritance"
    X_LINKED_DOMINANT_INHERITANCE = "X-linked dominant inheritance"
    Y_LINKED_INHERITANCE = "Y-linked inheritance"
    OTHER = "Other"
    X_LINKED_INHERITANCE = "X-linked inheritance"
    CODOMINANT = "Codominant"
    DEMIDOMINANT_INHERITANCE = "Semidominant inheritance"
    AUTOSOMAL_UNKNOWN = "Autosomal unknown"
    AUTOSOMAL_DOMINANT_INHERITANCE_WITH_MATERNAL_IMPRINTING = (
        "Autosomal dominant inheritance with maternal imprinting"
    )
    AUTOSOMAL_DOMINANT_INHERITANCE_WITH_PATERNAL_IMPRINTING = (
        "Autosomal dominant inheritance with paternal imprinting"
    )
    MULTIFACTORIAL_INHERITANCE = "Multifactorial inheritance"
    UNKNOWN_MECHANISM = "Unknown mechanism"
    OLIGOGENIC_INHERITANCE = "Oligogenic inheritance"


class RecordStatus(Enum):
    NOVEL = "novel"
    UPDATE = "update"


class ReleaseStatus(Enum):
    PUBLIC = "public"
    HOLD_UNTIL_PUBLISHED = "hold until published"


class SubmissionClinvarDeletionAccessionSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    accession: str
    reason: typing.Optional[str] = None


class SubmissionClinvarDeletion(BaseModel):
    model_config = ConfigDict(frozen=True)

    accessionSet: typing.List[SubmissionClinvarDeletionAccessionSet]


class SubmissionChromosomeCoordinates(BaseModel):
    model_config = ConfigDict(frozen=True)

    alternateAllele: typing.Optional[str] = None
    accession: typing.Optional[str] = None
    assembly: typing.Optional[Assembly] = None
    chromosome: typing.Optional[Chromosome] = None
    innerStart: typing.Optional[int] = None
    innerStop: typing.Optional[int] = None
    outerStart: typing.Optional[int] = None
    outerStop: typing.Optional[int] = None
    referenceAllele: typing.Optional[str] = None
    start: typing.Optional[int] = None
    stop: typing.Optional[int] = None
    variantLength: typing.Optional[int] = None


class SubmissionVariantGene(BaseModel):
    model_config = ConfigDict(frozen=True)

    id: typing.Optional[int] = None
    symbol: typing.Optional[str] = None


class SubmissionVariant(BaseModel):
    model_config = ConfigDict(frozen=True)

    chromosomeCoordinates: typing.Optional[SubmissionChromosomeCoordinates] = None
    copyNumber: typing.Optional[str] = None
    gene: typing.Optional[typing.List[SubmissionVariantGene]] = None
    hgvs: typing.Optional[str] = None
    referenceCopyNumber: typing.Optional[int] = None
    variantType: typing.Optional[VariantType] = None


class SubmissionVariantSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    variant: typing.List[SubmissionVariant]


class SubmissionPhaseUnknownSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    hgvs: str
    variants: typing.List[SubmissionVariant]


class SubmissionClinicalFeature(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinicalFeaturesAffectedStatus: ClinicalFeaturesAffectedStatus
    db: typing.Optional[ClinicalFeaturesDb] = None
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None


class SubmissionObservedIn(BaseModel):
    model_config = ConfigDict(frozen=True)

    affectedStatus: AffectedStatus
    alleleOrigin: AlleleOrigin
    collectionMethod: CollectionMethod
    clinicalFeatures: typing.Optional[typing.List[SubmissionClinicalFeature]] = None
    clinicalFeaturesComment: typing.Optional[str] = None
    numberOfIndividuals: typing.Optional[int] = None
    structVarMethodType: typing.Optional[StructVarMethodType] = None


class SubmissionHaplotypeSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    hgvs: str
    variants: typing.List[SubmissionVariant]
    starAlleleName: typing.Optional[str] = None


class SubmissionDistinctChromosomesSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    hgvs: str
    #: Hast at least two elements
    variants: typing.List[SubmissionVariant]


class SubmissionHaplotypeSets(BaseModel):
    model_config = ConfigDict(frozen=True)

    haplotypeSet: typing.Optional[SubmissionHaplotypeSet] = None
    haplotypeSingleVariantSet: typing.Optional[SubmissionHaplotypeSet] = None


class SubmissionDiplotypeSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    haplotypeSets: typing.List[SubmissionHaplotypeSets]
    hgvs: str
    starAlleleName: typing.Optional[str] = None


class SubmissionCitation(BaseModel):
    model_config = ConfigDict(frozen=True)

    db: typing.Optional[CitationDb] = None
    id: typing.Optional[str] = None
    url: typing.Optional[str] = None


class SubmissionAssertionCriteria(BaseModel):
    model_config = ConfigDict(frozen=True)

    db: typing.Optional[CitationDb] = None
    id: typing.Optional[str] = None
    url: typing.Optional[str] = None


class SubmissionCondition(BaseModel):
    model_config = ConfigDict(frozen=True)

    db: typing.Optional[ConditionDb] = None
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None


class SubmissionDrugResponse(BaseModel):
    model_config = ConfigDict(frozen=True)

    db: typing.Optional[ConditionDb] = None
    drugName: typing.Optional[str] = None
    id: typing.Optional[str] = None
    condition: typing.Optional[typing.List[SubmissionCondition]] = None


class SubmissionConditionSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: typing.Optional[typing.List[SubmissionCondition]] = None
    drugResponse: typing.Optional[typing.List[SubmissionDrugResponse]] = None
    multipleConditionExplanation: typing.Optional[MultipleConditionExplanation] = None


class SubmissionCompoundHeterozygoteSetVariantSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    variantSet: typing.Optional[SubmissionVariantSet] = None


class SubmissionCompoundHeterozygoteSet(BaseModel):
    model_config = ConfigDict(frozen=True)

    hgvs: str
    # Must have two entries
    variantSets: typing.List[SubmissionCompoundHeterozygoteSetVariantSet]


class SubmissionClinicalSignificance(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinicalSignificanceDescription: ClinicalSignificanceDescription
    citation: typing.Optional[typing.List[SubmissionCitation]] = None
    comment: typing.Optional[str] = None
    customAssertionScore: typing.Optional[float] = None
    dateLastEvaluated: typing.Optional[str] = None
    explanationOfDrugResponse: typing.Optional[str] = None
    explanationOfOtherClinicalSignificance: typing.Optional[str] = None
    modeOfInheritance: typing.Optional[ModeOfInheritance] = None


class SubmissionClinvarSubmission(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinicalSignificance: SubmissionClinicalSignificance
    conditionSet: SubmissionConditionSet
    observedIn: typing.List[SubmissionObservedIn]
    recordStatus: RecordStatus
    clinvarAccession: typing.Optional[str] = None
    compoundHeterozygoteSet: typing.Optional[SubmissionCompoundHeterozygoteSet] = None
    diplotypeSet: typing.Optional[SubmissionDiplotypeSet] = None
    distinctChromosomesSet: typing.Optional[SubmissionDistinctChromosomesSet] = None
    #: Has at least two elements in `variants`
    haplotypeSet: typing.Optional[SubmissionHaplotypeSet] = None
    #: Has exactly one elements in `variants`
    haplotypeSingleVariantSet: typing.Optional[SubmissionHaplotypeSet] = None
    localID: typing.Optional[str] = None
    localKey: typing.Optional[str] = None
    phaseUnknownSet: typing.Optional[SubmissionPhaseUnknownSet] = None
    variantSet: typing.Optional[SubmissionVariantSet] = None


class SubmissionContainer(BaseModel):
    """Representation of the container for a submission."""

    model_config = ConfigDict(frozen=True)

    assertionCriteria: typing.Optional[SubmissionAssertionCriteria] = None
    behalfOrgID: typing.Optional[int] = None
    clinvarDeletion: typing.Optional[SubmissionClinvarDeletion] = None
    clinvarSubmission: typing.Optional[typing.List[SubmissionClinvarSubmission]] = None
    clinvarSubmissionReleaseStatus: typing.Optional[ReleaseStatus] = None
    submissionName: typing.Optional[str] = None



import typing


def clean_for_json(
    value: typing.Union[
        bool, int, float, typing.List[typing.Any], None, typing.Dict[str, typing.Any]
    ]
) -> typing.Union[bool, int, float, typing.List[typing.Any], None, typing.Dict[str, typing.Any]]:
    """Clean the given value for JSON submission."""
    if isinstance(value, dict):
        return {k: clean_for_json(v) for k, v in value.items() if v is not None}
    elif isinstance(value, list):
        return [clean_for_json(elem) for elem in value]
    else:
        return value



"""Management of batches."""

import datetime
import pathlib
import typing

from logzero import logger
from tabulate import tabulate

from clinvar_api import client, models
from clinvar_this import config, exceptions
from clinvar_this.io import tsv


def get_share_dir():
    """Shared files directory"""
    return pathlib.Path.home() / ".local" / "share" / "clinvar-this"


#: Format string
FORMAT_STR = "%Y%m%d%H%M%S"


def _list_get_batches(share_dir: pathlib.Path):
    if not share_dir.exists():
        return []
    else:
        return [share_dir / path for path in share_dir.glob("*") if path.is_dir()]


def list_(config: config.Config):
    """List batches to stdout."""
    share_dir = get_share_dir()
    print(f"Listing batches at {share_dir}/{config.profile}")
    _ = config
    paths = _list_get_batches(share_dir / config.profile)
    if not paths:
        table = [["-- NO BATCHES YET --"]]
        print(tabulate(table))
    else:
        table = [[path.name] for path in sorted(paths)]
        print(tabulate(table, headers=["path"], tablefmt="grid"))


def gen_name(config: config.Config) -> str:
    """Generate batch name that does not exist yet."""
    base = datetime.date.today().strftime("%Y-%m-%d")
    for i in range(1000):
        dirname = "%s-%03d" % (base, i)
        if not (get_share_dir() / config.profile / dirname).exists():
            return dirname
    else:  # pragma: no cover
        raise exceptions.IOException("Could not generate batch name")


def _write_payload(submission_container: models.SubmissionContainer, profile: str, name: str):
    """Write out payload to a new JSON file."""
    # Create directory for batch.
    batch_dir = get_share_dir() / profile / name
    batch_dir.mkdir(exist_ok=True, parents=True)
    # Write out payload.
    timestamp = datetime.datetime.now().strftime(FORMAT_STR)
    payload_path = batch_dir / f"payload.{timestamp}.json"
    payload_json = submission_container.model_dump_json(indent=2)
    with payload_path.open("wt") as outputf:
        outputf.write(payload_json)
        outputf.write("\n")


def _merge_submission_container(
    base: models.SubmissionContainer,
    patch: models.SubmissionContainer,
) -> models.SubmissionContainer:
    """Update base submission container with new one.

    The following attributes will be copied from ``patch`` to ``base``:

    - mode of inheritance
    - clinical significance
        - clinical significance description (ACMG grading)
        - condition
    """
    logger.info("Merging submission information...")

    def merge_submission(
        base: models.SubmissionClinvarSubmission,
        patch: models.SubmissionClinvarSubmission,
    ) -> models.SubmissionClinvarSubmission:
        clinvar_accession = base.clinvar_accession or patch.clinvar_accession
        return base.model_copy(
            update={
                "clinvar_accession": clinvar_accession,
                "condition_set": patch.condition_set,
                "clinical_significance": patch.clinical_significance,
                "observed_in": patch.observed_in,
            }
        )

    patch_clinvar_submission = {
        clinvar_submission.local_key: clinvar_submission
        for clinvar_submission in (patch.clinvar_submission or [])
    }
    clinvar_submissions = []
    for submission in base.clinvar_submission or []:
        if submission.local_key in patch_clinvar_submission:
            clinvar_submissions.append(
                merge_submission(submission, patch_clinvar_submission[submission.local_key])
            )
        else:
            clinvar_submissions.append(submission)
    result = base.model_copy(update={"clinvar_submission": clinvar_submissions})
    logger.info("... done merging submission information")
    return result


def import_(config: config.Config, name: str, path: str, metadata: typing.Tuple[str, ...]):
    """Import the data file at ``path`` into the batch of name ``name``."""
    existing_payloads = list((get_share_dir() / config.profile / name).glob("payload.*.json"))
    if existing_payloads:
        logger.info("Loading existing payload for later merging with new one")
        previous_submission_container = _load_latest_payload(config.profile, name)
    else:
        logger.info("Creating new payload only")
        previous_submission_container = None
    if path.endswith(".tsv") or path.endswith(".txt"):
        tsv_type = tsv.guess_tsv_type(path)
        if tsv_type in (tsv.TsvType.SEQ_VAR, tsv.TsvType.STRUC_VAR):
            batch_metadata = tsv.batch_metadata_from_mapping(metadata, use_defaults=True)
            if tsv_type == tsv.TsvType.SEQ_VAR:
                new_submission_container = tsv.seq_var_tsv_records_to_submission_container(
                    tsv.read_seq_var_tsv(path=path), batch_metadata
                )
            else:  # tsv_type == tsv.TsvType.STRUC_VAR
                new_submission_container = tsv.struc_var_tsv_records_to_submission_container(
                    tsv.read_struc_var_tsv(path=path), batch_metadata
                )
            if previous_submission_container:
                submission_container = _merge_submission_container(
                    base=previous_submission_container,
                    patch=new_submission_container,
                )
            else:
                submission_container = new_submission_container
        else:
            raise exceptions.IOException(f"Could not guess TSV file type from header for {path}")
        _write_payload(submission_container, config.profile, name)
    else:  # pragma: no cover
        raise exceptions.IOException(f"File extension of {path} cannot be handled.")


def _load_latest_payload(profile: str, name: str):
    submission_path = get_share_dir() / profile / name
    payload_paths = list(sorted(submission_path.glob("payload.*.json")))
    if not payload_paths:  # pragma: no cover
        raise exceptions.ClinvarThisException(f"Found no payload JSON file at {submission_path}")

    payload_path = submission_path / payload_paths[-1]
    with payload_path.open("rt") as inputf:
        payload_json = inputf.read()
    return models.SubmissionContainer.model_validate_json(payload_json)


def export(
    config: config.Config, name: str, path: str, force: bool = False, struc_var: bool = False
):
    """Export the batch with the given ``name`` to the file at ``path``."""
    if pathlib.Path(path).exists() and not force:
        raise exceptions.IOException(
            f"File at output path {path} already exists. Use --force to overwrite."
        )
    if path.endswith(".tsv") or path.endswith(".txt"):
        payload = _load_latest_payload(config.profile, name)
        if struc_var:
            tsv.write_struc_var_tsv(
                tsv_records=tsv.submission_container_to_struc_var_tsv_records(payload), path=path
            )
        else:
            tsv.write_seq_var_tsv(
                tsv_records=tsv.submission_container_to_seq_var_tsv_records(payload), path=path
            )
    else:  # pragma: no cover
        raise exceptions.IOException(f"File extension of {path} cannot be handled.")


def update_metadata(config: config.Config, name: str, metadata: typing.Tuple[str, ...]):
    """Update the batch' meta data."""
    batch_metadata = tsv.batch_metadata_from_mapping(metadata, use_defaults=False)
    _ = batch_metadata


def submit(config: config.Config, name: str, *, use_testing: bool = False, dry_run: bool = False):
    """Submit the batch to ClinVar."""
    if not config.auth_token:  # pragma: no cover
        raise exceptions.ConfigException("auth_token not configured")

    client_obj = client.Client(
        client.Config(
            auth_token=config.auth_token,
            use_testing=use_testing,
            use_dryrun=dry_run,
            verify_ssl=config.verify_ssl,
        )
    )

    payload = _load_latest_payload(config.profile, name)

    logger.info("Initiating submission to ClinVar API")
    client_res = client_obj.submit_data(payload)

    # Terminate earlyier in dry-run mode.
    if dry_run:
        logger.info("In dry-run mode, not writing out response.")
        return

    timestamp = datetime.datetime.now().strftime(FORMAT_STR)
    response_path = (
        get_share_dir() / config.profile / name / f"submission-response.{timestamp}.json"
    )

    logger.info("Writing out server response to %s", response_path)
    with response_path.open("wt") as outputf:
        print(client_res.model_dump_json(), file=outputf)
    logger.info(
        "The ClinVar API has accepted your submission and will perform additional checks in the background."
    )
    logger.info(
        (
            "The next step is to run ``clinvar-this batch retrieve %s`` and wait until you get a "
            "final success or error response"
        ),
        name,
    )
    logger.info("All done. Have a nice day!")


def _retrieve_store_response(
    config: config.Config, name: str, status_result: client.RetrieveStatusResult
):
    """Store information from the retrieve status result into a new payload JSON file.

    Currently, this is the clinvar accession only.
    """
    logger.debug("Updating local payload from retrieve status response ...")
    logger.debug("Obtaining local key to accession map")
    local_key_to_accession = {}
    local_id_to_error = {}
    for summary_response in status_result.summaries.values():
        submissions = summary_response.submissions or []
        for submission in submissions:
            local_key_to_accession[
                submission.identifiers.local_key or submission.identifiers.clinvar_local_key
            ] = submission.identifiers.clinvar_accession
            errors = [
                error_inner.user_message
                for error_outer in (submission.errors or [])
                for error_inner in (error_outer.output.errors or [])
            ]
            local_id_to_error[submission.identifiers.local_id] = "; ".join(errors)
    logger.debug("Update map is %s", local_key_to_accession)
    logger.debug("Loading latest payload")
    payload = _load_latest_payload(config.profile, name)
    logger.debug("Updating local payload")
    clinvar_submission = [
        submission.model_copy(
            update={
                "clinvar_accession": local_key_to_accession.get(
                    submission.local_key, submission.clinvar_accession
                ),
                "record_status": models.RecordStatus.UPDATE
                if submission.local_key in local_key_to_accession
                else models.RecordStatus.NOVEL,
                "extra_data": {
                    **(submission.extra_data or {}),
                    "error_msg": local_id_to_error.get(submission.local_id, ""),
                },
            }
        )
        for submission in payload.clinvar_submission
    ]
    updated_payload = payload.model_copy(update={"clinvar_submission": clinvar_submission})
    logger.debug("Write out updated payload")
    _write_payload(updated_payload, config.profile, name)
    logger.debug("... done updating local payload from retrieve status response")


def retrieve(config: config.Config, name: str, *, use_testing: bool = False):
    """Retrieve current processing status from ClinVar."""
    client_obj = client.Client(client.Config(auth_token=config.auth_token, use_testing=use_testing))

    submission_path = get_share_dir() / config.profile / name
    submission_response_paths = list(sorted(submission_path.glob("submission-response.*.json")))

    if not submission_path.exists():  # pragma: no cover
        raise exceptions.ClinvarThisException(f"Submission does not exist at {submission_path}")
    elif not submission_response_paths:  # pragma: no cover
        raise exceptions.ClinvarThisException(
            f"Submission not submitted? No submission response at {submission_path}"
        )

    submission_response_path = submission_path / submission_response_paths[-1]
    logger.info("Loading response from %s", submission_response_path)
    with submission_response_path.open("rb") as inputf:
        created = models.Created.model_validate_json(inputf.read())
    logger.info("Submission ID is %s", created.id)

    logger.info("Initiating fetching of status from ClinVar API")
    status_result = client_obj.retrieve_status(created.id)
    timestamp = datetime.datetime.now().strftime(FORMAT_STR)
    retrieve_response_path = submission_path / f"retrieve-response.{timestamp}.json"
    logger.debug("Writing out response to %s", retrieve_response_path)
    with retrieve_response_path.open("wt") as outputf:
        print(status_result.model_dump_json(indent=2), file=outputf)

    status_str = status_result.status.actions[0].status
    if status_str in ["submitted", "processing"]:
        logger.info(f"Status is {status_str}, be patient and check back in a while...")
        logger.info(
            "Submissions with errors tend to fail quickly while successful submissions tend to take a while."
        )
    elif status_str == "processed":
        logger.info("Submission has been processed successfully")
        logger.info("Will now update local information from response...")
        _retrieve_store_response(config, name, status_result)
        logger.info("... done updating local information from response")
    elif status_str == "error":
        logger.error("There were errors in your submission")
        logger.info("Check the file %s for details", retrieve_response_path)
        logger.info("Will now update local information from response...")
        _retrieve_store_response(config, name, status_result)
        logger.info("... done updating local information from response")
    else:  # pragma: no cover
        logger.error("Status is %s and clinvar-this does not know how to handle this yet!")
        raise exceptions.ClinvarThisException(f"Unknown status {status_str}")



"""REST API client code for communicating with server endpoints."""

import asyncio
import contextlib
import json
import typing

import httpx
from jsonschema import ValidationError
from logzero import logger
from pydantic import BaseModel, SecretStr
from pydantic.config import ConfigDict

from clinvar_api import common, exceptions, models, msg, schemas

#: URL of the server endpoint (non-test/production).
ENDPOINT_URL_PROD = "https://submit.ncbi.nlm.nih.gov/api/v1/submissions/"

#: URL of the test endpoint.
ENDPOINT_URL_TEST = "https://submit.ncbi.nlm.nih.gov/apitest/v1/submissions/"

#: URL suffix for enabling dry-run.
SUFFIX_DRYRUN = "?dry-run=true"


class Config(BaseModel):
    """Configuration for the ``Client`` class."""

    model_config = ConfigDict(frozen=True)

    #: Token to use for authentication.
    auth_token: SecretStr

    #: Whether to use the test endpoint.
    use_testing: bool = False

    #: Whether to use dry running.
    use_dryrun: bool = False

    #: Whether to validate submission payload before posting.
    presubmission_validation: bool = True

    #: Whether or not to verify SSL on submission.
    verify_ssl: bool = True


class _SubmitData:
    """Helper class to reduce redundancy betwee sync/async `sumbit_data`."""

    def __init__(self, submission_container: models.SubmissionContainer, config: Config):
        self.submission_container = submission_container
        self.config = config

        self.url: str = ""
        self.headers: typing.Dict[str, str] = {}
        self.post_data: typing.Dict[str, typing.Any] = {}

    def before_post(self) -> typing.Tuple[str, typing.Dict[str, str], typing.Dict[str, typing.Any]]:
        logger.info("Submitting with config %s", self.config)

        url_prefix = ENDPOINT_URL_TEST if self.config.use_testing else ENDPOINT_URL_PROD
        url_suffix = SUFFIX_DRYRUN if self.config.use_dryrun else ""
        url = f"{url_prefix}{url_suffix}"
        logger.debug("Will submit to URL %s", url)
        headers = {
            "SP-API-KEY": self.config.auth_token.get_secret_value(),
        }

        payload = self.submission_container.to_msg().model_dump(mode="json")
        logger.debug("Payload data is %s", json.dumps(payload, indent=2))
        cleaned_payload = common.clean_for_json(payload)
        logger.debug("Cleaned payload data is %s", json.dumps(cleaned_payload, indent=2))
        if self.config.presubmission_validation:
            logger.info("Validating payload...")
            schemas.validate_submission_payload(cleaned_payload)
            logger.info("... done validating payload")
        else:
            logger.info("Configured to NOT validate payload before submission")

        post_data = {
            "actions": [
                {"type": "AddData", "targetDb": "clinvar", "data": {"content": cleaned_payload}}
            ]
        }
        logger.debug("Overall POST payload is %s", post_data)
        return url, headers, post_data

    def after_post(self, response: httpx.Response):
        if httpx.codes.is_success(response.status_code):
            logger.info("API returned OK - %s", response.status_code)
            if response.status_code == 204:  # no content, on dry-run
                logger.info("Server returned '204: No Content', constructing fake created message.")
                return models.Created(id="--NONE--dry-run-result--")
            else:
                created_msg = msg.Created.model_validate_json(response.content)
                return models.Created.from_msg(created_msg)
        else:
            logger.warning("API returned an error - %s", response.status_code)
            error_msg = msg.Error.model_validate_json(response.content)
            error_obj = models.Error.from_msg(error_msg)
            logger.debug("Full server response is %s", response.json())
            if hasattr(error_obj, "errors"):
                raise exceptions.SubmissionFailed(
                    f"ClinVar submission failed: {error_obj.message}, errors: {error_obj.errors}"
                )
            else:
                raise exceptions.SubmissionFailed(f"ClinVar submission failed: {error_obj.message}")


def submit_data(submission_container: models.SubmissionContainer, config: Config) -> models.Created:
    """Submit new data to ClinVar API (sync).

    :param submission_container: The submission data.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.SubmissionFailed: on problems with the submission.
    """
    helper = _SubmitData(submission_container, config)
    url, headers, post_data = helper.before_post()
    response = httpx.post(url, headers=headers, json=post_data, verify=config.verify_ssl)
    return helper.after_post(response)


async def async_submit_data(
    submission_container: models.SubmissionContainer, config: Config
) -> models.Created:
    """Submit new data to ClinVar API via async API (async).

    :param submission_container: The submission data.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.SubmissionFailed: on problems with the submission.
    """
    helper = _SubmitData(submission_container, config)
    url, headers, post_data = helper.before_post()
    async with httpx.AsyncClient(verify=config.verify_ssl) as client:
        response = await client.post(url, headers=headers, json=post_data)
    return helper.after_post(response)


class RetrieveStatusResult(BaseModel):
    """Result type for ``retrieve_status`` function."""

    model_config = ConfigDict(frozen=True)

    #: The submission status.
    status: models.SubmissionStatus
    #: A dict mapping file URLs to the parsed ``Sum``.
    summaries: typing.Dict[str, models.SummaryResponse]


def _handle_retrieved_status_summaries(
    response: httpx.Response, validate_response_json: bool = True
) -> models.SummaryResponse:
    """Handle retrieved status summary from the given URL."""
    if httpx.codes.is_success(response.status_code):
        response_json = response.json()
        if validate_response_json:
            logger.debug("Validating status summary response ...")
            try:
                schemas.validate_status_summary(response_json)
            except ValidationError as e:
                logger.warning("Response summary validation JSON is invalid: %s", e)
            logger.debug("... done validating status summary response")
        sr_msg = msg.SummaryResponse.model_validate_json(response.content)
        return models.SummaryResponse.from_msg(sr_msg)
    else:
        raise exceptions.QueryFailed(f"Could not perform query: {response.status_code}")


class _RetrieveStatus:
    """Helper class to reduce redundancy betwee sync/async `retrieve_status`."""

    def __init__(self, submission_id: str, config: Config):
        self.submission_id = submission_id
        self.config = config

    def before_first_get(self) -> typing.Tuple[str, typing.Dict[str, str]]:
        url_prefix = ENDPOINT_URL_TEST if self.config.use_testing else ENDPOINT_URL_PROD
        url_suffix = SUFFIX_DRYRUN if self.config.use_dryrun else ""
        url = f"{url_prefix}{self.submission_id}/actions/{url_suffix}"
        headers = {
            "SP-API-KEY": self.config.auth_token.get_secret_value(),
        }
        logger.debug("Will query URL %s", url)
        return url, headers

    def after_first_get_failure(self, response: httpx.Response):
        logger.info("API returned an error %s", response.status_code, response)
        response_json = response.json()
        raise exceptions.QueryFailed(f"ClinVar query failed: {response_json}")

    def after_first_get_success(
        self, response: httpx.Response
    ) -> typing.Tuple[typing.List[str], models.SubmissionStatus]:
        logger.info("API returned OK - %s", response.status_code)
        logger.debug("Structuring response ...")
        status_msg = msg.SubmissionStatus.model_validate_json(response.content)
        logger.debug(
            "structured response is %s",
            status_msg.model_dump_json(indent=2),
        )
        logger.debug("... done structuring response")
        status_obj = models.SubmissionStatus.from_msg(status_msg)
        logger.info(
            "Attempting to fetch %d status summary files...",
            len(
                [
                    None
                    for action in status_obj.actions
                    for action_response in action.responses
                    for _ in action_response.files
                ]
            ),
        )
        more_urls: typing.List[str] = []
        for action in status_obj.actions:
            for action_response in action.responses:
                for file_ in action_response.files:
                    more_urls.append(file_.url)
        return more_urls, status_obj

    def after_get_more_urls(
        self, status_obj: models.SubmissionStatus, more_results: typing.Dict[str, httpx.Response]
    ) -> RetrieveStatusResult:
        summaries = {}
        for url, response in more_results.items():
            summaries[url] = _handle_retrieved_status_summaries(response)
        logger.info("... done fetching status summary files")
        return RetrieveStatusResult(status=status_obj, summaries=summaries)


def retrieve_status(
    submission_id: str,
    config: Config,
) -> RetrieveStatusResult:
    """Retrieve submission status from API.

    :param submission_id: The identifier of the submission as returned earlier from API.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.QueryFailed: on problems with the communication to the server.
    """
    helper = _RetrieveStatus(submission_id, config)
    url, headers = helper.before_first_get()
    response = httpx.get(url, headers=headers)
    if httpx.codes.is_success(response.status_code):
        more_urls, status_obj = helper.after_first_get_success(response)

        more_results = {}
        for url in more_urls:
            logger.info(" - fetching %s", url)
            more_results[url] = httpx.get(url)

        return helper.after_get_more_urls(status_obj, more_results)
    else:
        return helper.after_first_get_failure(response)


async def async_retrieve_status(
    submission_id: str,
    config: Config,
) -> RetrieveStatusResult:
    """Retrieve submission status from API.

    :param submission_id: The identifier of the submission as returned earlier from API.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.QueryFailed: on problems with the communication to the server.
    """
    helper = _RetrieveStatus(submission_id, config)
    url, headers = helper.before_first_get()
    async with httpx.AsyncClient(verify=config.verify_ssl) as client:
        response = await client.get(url, headers=headers)
    if httpx.codes.is_success(response.status_code):
        more_urls, status_obj = helper.after_first_get_success(response)

        async with contextlib.AsyncExitStack() as stack:
            tasks: typing.Dict[str, typing.Awaitable[httpx.Response]] = {}
            for url in more_urls:
                logger.info(" - fetching %s", url)
                client = await stack.enter_async_context(
                    httpx.AsyncClient(verify=config.verify_ssl)
                )
                tasks[url] = client.get(url)

            more_results: typing.Dict[str, httpx.Response] = dict(
                zip(tasks.keys(), await asyncio.gather(*tasks.values()))
            )

        return helper.after_get_more_urls(status_obj, more_results)
    else:
        return helper.after_first_get_failure(response)


class Client:
    """NCBI ClinVar REST API client (sync)."""

    def __init__(self, config: Config):
        self.config = config

    def submit_data(self, payload: models.SubmissionContainer) -> models.Created:
        """Submit new data to ClinVar API.

        :param payload: The submission data.
        :return: The information about the created submission.
        :raises exceptions.SubmissionFailed: on problems with the submission.
        """
        return submit_data(payload, self.config)

    def retrieve_status(self, submission_id: str) -> RetrieveStatusResult:
        """Retrieve submission status from API.

        :param submission_id: The identifier of the submission as returned earlier from API.
        :return: The information about the created submission.
        :raises exceptions.QueryFailed: on problems with the communication to the server.
        """
        return retrieve_status(submission_id, self.config)


class AsyncClient:
    """NCBI ClinVar REST API client (async)."""

    def __init__(self, config: Config):
        self.config = config

    async def submit_data(self, payload: models.SubmissionContainer) -> models.Created:
        """Submit new data to ClinVar API.

        :param payload: The submission data.
        :return: The information about the created submission.
        :raises exceptions.SubmissionFailed: on problems with the submission.
        """
        return await async_submit_data(payload, self.config)

    async def retrieve_status(self, submission_id: str) -> RetrieveStatusResult:
        """Retrieve submission status from API.

        :param submission_id: The identifier of the submission as returned earlier from API.
        :return: The information about the created submission.
        :raises exceptions.QueryFailed: on problems with the communication to the server.
        """
        return await async_retrieve_status(submission_id, self.config)



"""Generate JSONL with gene to phenotype/disease links"""

import gzip
import json
import typing

from pydantic import BaseModel, ConfigDict
import tqdm

from clinvar_data import models


class GenePhenotypeRecord(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: RCV accession
    rcv: str
    #: RCV accession version
    rcv_version: int
    #: SCV accession
    scv: str
    #: SCV accession version
    scv_version: int
    #: Clinical significance
    clinsig: typing.Optional[models.ClinicalSignificanceDescription]
    #: Submitter
    submitter: typing.Optional[str]
    #: Gene HGNC ID
    hgnc_ids: typing.List[str]
    #: Linked OMIM terms
    omim_terms: typing.List[str]
    #: Linked MONDO terms
    mondo_terms: typing.List[str]
    #: Linked HPO terms
    hpo_terms: typing.List[str]


def clean_omim(terms: typing.Set[str]) -> typing.Set[str]:
    return {term.split(".")[0] for term in terms}


class TermCollector:
    def __init__(self):
        self.omim_terms = set()
        self.mondo_terms = set()
        self.hpo_terms = set()

    def add_xref(self, xref: models.Xref):
        if xref.db == "OMIM":
            self.omim_terms.add(xref.id)
        elif xref.db in ("Human Phenotype Ontology", "HP"):
            self.hpo_terms.add(xref.id)
        elif xref.db == "MONDO":
            self.mondo_terms.add(xref.id)

    def has_link(self) -> bool:
        return self.hpo_terms or self.omim_terms or self.mondo_terms


def run_report(path_input: str, path_output: str, needs_hpo_terms: bool = True):
    """Read in file at path_input and generate link records to path_output."""

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    if path_output.endswith(".gz"):
        outputf = gzip.open(path_output, "wt")
    else:
        outputf = open(path_output, "wt")

    with inputf, outputf:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            clinvar_set = models.ClinVarSet.model_validate_json(line)
            rca = clinvar_set.reference_clinvar_assertion
            rcv = rca.clinvar_accession.acc
            rcv_version = rca.clinvar_accession.version

            hgnc_ids = set()
            if rca.measures:
                for measure in rca.measures.measures or []:
                    for measure_relationship in measure.measure_relationship:
                        hgnc_ids |= {
                            xref.id for xref in measure_relationship.xrefs if xref.db == "HGNC"
                        }

            for ca in clinvar_set.clinvar_assertions or []:
                clinsig = None
                for clinical_significance in ca.clinical_significance:
                    for description in clinical_significance.descriptions:
                        if description in (
                            models.ClinicalSignificanceDescription.PATHOGENIC,
                            models.ClinicalSignificanceDescription.LIKELY_PATHOGENIC,
                        ):
                            clinsig = description
                            break

                scv = ca.clinvar_accession.acc
                scv_version = ca.clinvar_accession.version

                terms = TermCollector()
                for observed_in in ca.observed_in or []:
                    if observed_in.traits:
                        for trait in observed_in.traits.traits or []:
                            for xref in trait.xrefs:
                                terms.add_xref(xref)

                if needs_hpo_terms and not terms.hpo_terms or not terms.has_link():
                    continue

                record = GenePhenotypeRecord(
                    rcv=rcv,
                    rcv_version=rcv_version,
                    scv=scv,
                    scv_version=scv_version,
                    clinsig=clinsig,
                    submitter=ca.submission_id.submitter,
                    hgnc_ids=list(sorted(hgnc_ids)),
                    omim_terms=list(sorted(clean_omim(terms.omim_terms))),
                    mondo_terms=list(sorted(terms.mondo_terms)),
                    hpo_terms=list(sorted(terms.hpo_terms)),
                )

                print(json.dumps(record.model_dump(mode="json")), file=outputf)



"""Data structures for internal representation of query response."""

import datetime
import typing

from pydantic import BaseModel
from pydantic.config import ConfigDict

from clinvar_api import msg
from clinvar_api.msg import ErrorCode


class SubmissionStatusFile(BaseModel):
    model_config = ConfigDict(frozen=True)

    url: str

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusFile):
        return SubmissionStatusFile(url=other.url)


class SubmissionStatusObjectContent(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: Processing status
    clinvar_processing_status: str
    #: Release status
    clinvar_release_status: str

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusObjectContent):
        return SubmissionStatusObjectContent(
            clinvar_processing_status=other.clinvarProcessingStatus,
            clinvar_release_status=other.clinvarReleaseStatus,
        )


class SubmissionStatusObject(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: Optional object accession.
    accession: typing.Optional[str]
    #: Object content.
    content: SubmissionStatusObjectContent
    #: Target database, usually "clinvar" per the docs.
    target_db: str

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusObject):
        return SubmissionStatusObject(
            accession=other.accession,
            content=SubmissionStatusObjectContent.from_msg(other.content),
            target_db=other.targetDb,
        )


class SubmissionStatusResponseMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: The error code.
    error_code: typing.Optional[ErrorCode]
    #: The message severity.
    severity: str
    #: The message text.
    text: str

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusResponseMessage):
        return SubmissionStatusResponseMessage(
            error_code=other.errorCode, severity=other.severity, text=other.text
        )


class SubmissionStatusResponse(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: Status, one of "processing", "processed", "error",
    status: str
    #: Files in the response.
    files: typing.List[SubmissionStatusFile]
    #: Message
    message: typing.Optional[SubmissionStatusResponseMessage]
    #: Objects
    objects: typing.List[SubmissionStatusObject]

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusResponse):
        if other.message:
            message = SubmissionStatusResponseMessage.from_msg(other.message)
        else:
            message = None
        return SubmissionStatusResponse(
            status=other.status,
            files=[SubmissionStatusFile.from_msg(f) for f in other.files],
            message=message,
            objects=list(map(SubmissionStatusObject.from_msg, other.objects)),
        )


class SubmissionStatusActions(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: Identifier of the submission
    id: str
    #: Entries in ``actions[*].responses``, only one entry per the docs.
    responses: typing.List[SubmissionStatusResponse]
    #: Status of the submission, one of "submitted", "processing", "processed", "error"
    status: str
    #: Target database, usually "clinvar"
    target_db: str
    #: Last updated time
    updated: datetime.datetime

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatusActions):
        return SubmissionStatusActions(
            id=other.id,
            responses=[SubmissionStatusResponse.from_msg(response) for response in other.responses],
            status=other.status,
            target_db=other.targetDb,
            updated=other.updated,
        )


class SubmissionStatus(BaseModel):
    """Internal submission status."""

    model_config = ConfigDict(frozen=True)

    #: The list of actions (one element only by the docs).
    actions: typing.List[SubmissionStatusActions]

    @classmethod
    def from_msg(cls, other: msg.SubmissionStatus):
        return SubmissionStatus(
            actions=[SubmissionStatusActions.from_msg(action) for action in other.actions]
        )



"""Data structures for submission response."""

import typing

from pydantic import BaseModel
from pydantic.config import ConfigDict

from clinvar_api import msg

# Import and re-use the enums. This makes coupling tighter than it would have to be but the
# trade-off of copy-pasting them with no influence on the interal record API makes that not
# worth it.
from clinvar_api.msg.sub_response import BatchProcessingStatus, BatchReleaseStatus


class Created(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: The submission ID.
    id: str

    @classmethod
    def from_msg(cls, other: msg.Created):
        return Created(id=other.id)


class Error(BaseModel):
    model_config = ConfigDict(frozen=True)

    #: The error response's message.
    message: str

    @classmethod
    def from_msg(cls, other: msg.Error):
        return Error(message=other.message)


class SummaryResponseErrorInput(BaseModel):
    model_config = ConfigDict(frozen=True)

    value: typing.Optional[str] = None
    field: typing.Optional[str] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseErrorInput):
        return SummaryResponseErrorInput(
            value=other.value,
            field=other.field,
        )


class SummaryResponseErrorOutputError(BaseModel):
    model_config = ConfigDict(frozen=True)

    user_message: str

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseErrorOutputError):
        return SummaryResponseErrorOutputError(
            user_message=other.userMessage,
        )


class SummaryResponseErrorOutput(BaseModel):
    model_config = ConfigDict(frozen=True)

    errors: typing.List[SummaryResponseErrorOutputError]

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseErrorOutput):
        return SummaryResponseErrorOutput(
            errors=[
                SummaryResponseErrorOutputError.from_msg(msg_error) for msg_error in other.errors
            ]
        )


class SummaryResponseError(BaseModel):
    model_config = ConfigDict(frozen=True)

    input: typing.List[SummaryResponseErrorInput]
    output: SummaryResponseErrorOutput

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseError):
        return SummaryResponseError(
            input=[SummaryResponseErrorInput.from_msg(msg_input) for msg_input in other.input],
            output=SummaryResponseErrorOutput.from_msg(other.output),
        )


class SummaryResponseDeletionIdentifier(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinvar_accession: str
    clinvar_local_key: typing.Optional[str] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseDeletionIdentifier):
        return SummaryResponseDeletionIdentifier(
            clinvar_accession=other.clinvarAccession,
            clinvar_local_key=other.clinvarLocalKey,
        )


class SummaryResponseDeletion(BaseModel):
    model_config = ConfigDict(frozen=True)

    identifiers: SummaryResponseDeletionIdentifier
    processing_status: str
    delete_date: typing.Optional[str] = None
    delete_status: typing.Optional[str] = None
    errors: typing.Optional[typing.List[SummaryResponseError]] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseDeletion):
        errors = None
        if other.errors:
            errors = [SummaryResponseError.from_msg(msg_error) for msg_error in other.errors]
        return SummaryResponseDeletion(
            identifiers=SummaryResponseDeletionIdentifier.from_msg(other.identifiers),
            processing_status=other.processingStatus,
            delete_date=other.deleteDate,
            delete_status=other.deleteStatus,
            errors=errors,
        )


class SummaryResponseSubmissionIdentifiers(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinvar_local_key: str
    clinvar_accession: typing.Optional[str] = None
    local_id: typing.Optional[str] = None
    local_key: typing.Optional[str] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseSubmissionIdentifiers):
        return SummaryResponseSubmissionIdentifiers(
            clinvar_local_key=other.clinvarLocalKey,
            clinvar_accession=other.clinvarAccession,
            local_id=other.localID,
            local_key=other.localKey,
        )


class SummaryResponseSubmission(BaseModel):
    model_config = ConfigDict(frozen=True)

    identifiers: SummaryResponseSubmissionIdentifiers
    processing_status: str
    clinvar_accession_version: typing.Optional[str] = None
    errors: typing.Optional[typing.List[SummaryResponseError]] = None
    release_date: typing.Optional[str] = None
    release_status: typing.Optional[str] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponseSubmission):
        errors = None
        if other.errors:
            errors = [SummaryResponseError.from_msg(msg_error) for msg_error in other.errors]
        return SummaryResponseSubmission(
            identifiers=SummaryResponseSubmissionIdentifiers.from_msg(
                other.identifiers,
            ),
            processing_status=other.processingStatus,
            clinvar_accession_version=other.clinvarAccessionVersion,
            errors=errors,
            release_date=other.releaseDate,
            release_status=other.releaseStatus,
        )


class SummaryResponse(BaseModel):
    """Represetation of server's response to a submission."""

    model_config = ConfigDict(frozen=True)

    batch_processing_status: BatchProcessingStatus
    batch_release_status: BatchReleaseStatus
    submission_date: str
    submission_name: str
    total_count: int
    total_errors: int
    total_public: int
    total_success: int
    deletions: typing.Optional[typing.List[SummaryResponseDeletion]] = None
    submissions: typing.Optional[typing.List[SummaryResponseSubmission]] = None
    total_delete_count: typing.Optional[int] = None
    total_deleted: typing.Optional[int] = None
    total_delete_errors: typing.Optional[int] = None
    total_delete_success: typing.Optional[int] = None

    @classmethod
    def from_msg(cls, other: msg.SummaryResponse):
        deletions = None
        if other.deletions:
            deletions = [
                SummaryResponseDeletion.from_msg(msg_deletion) for msg_deletion in other.deletions
            ]
        submissions = None
        if other.submissions:
            submissions = [
                SummaryResponseSubmission.from_msg(msg_deletion)
                for msg_deletion in other.submissions
            ]
        return SummaryResponse(
            batch_processing_status=other.batchProcessingStatus,
            batch_release_status=other.batchReleaseStatus,
            submission_date=other.submissionDate,
            submission_name=other.submissionName,
            total_count=other.totalCount,
            total_errors=other.totalErrors,
            total_public=other.totalPublic,
            total_success=other.totalSuccess,
            deletions=deletions,
            submissions=submissions,
            total_delete_count=other.totalDeleteCount,
            total_deleted=other.totalDeleted,
            total_delete_errors=other.totalDeleteErrors,
            total_delete_success=other.totalDeleteSuccess,
        )



"""Configuration management."""

import datetime
import pathlib
import sys

from pydantic import BaseModel, SecretStr
from pydantic.config import ConfigDict
import toml

from clinvar_this import exceptions


class Config(BaseModel):
    """Configuration for the ``clinvar-this`` app."""

    model_config = ConfigDict(frozen=True)

    #: The name of the profile.
    profile: str

    #: The authentication token to use in the API.
    auth_token: SecretStr

    #: Whether to verify SSL or not
    verify_ssl: bool = True


def load_config(profile: str = "default") -> Config:
    """Load configuration for the given profile.

    :params profile: The profile to load configuration for.
    :returns: The ``Config`` with the configuration.
    :raises exceptions.ConfigException: In the case of problems with configuration.
    """

    config_path = pathlib.Path.home() / ".config" / "clinvar-this" / "config.toml"

    if not config_path.exists():
        raise exceptions.ConfigFileMissingException(
            f"Configuration file {config_path} does not exist. Try `clinvar-this config set auth_token XXX`."
        )

    with config_path.open("rt") as configf:
        try:
            config_dict = toml.load(configf)
        except toml.TomlDecodeError as e:
            raise exceptions.ConfigException(
                f"Problem decoding configuration file {config_path}"
            ) from e

    return Config(profile=profile, auth_token=config_dict.get(profile, {}).get("auth_token"))


def save_config(config: Config, profile: str = "default"):
    """Save configuration to the given profile."""

    config_path = pathlib.Path.home() / ".config" / "clinvar-this" / "config.toml"

    if not config_path.parent.exists():
        config_path.parent.mkdir(parents=True)

    all_config = {}
    if config_path.exists():
        with config_path.open("rt") as configf:
            all_config = toml.load(configf)
        # create backup
        suffix = datetime.datetime.now().strftime("%Y%m%d-%H%M%S.%f")
        backup_path = config_path.parent / (config_path.name + f"~{suffix}")
        config_path.rename(backup_path)

    all_config.setdefault("default", {})
    all_config[profile] = {
        k: v for k, v in config.model_dump(mode="json").items() if k != "profile"
    }
    all_config[profile]["auth_token"] = config.auth_token.get_secret_value()

    with config_path.open("wt") as configf:
        toml.dump(all_config, configf)


def dump_config(outf=None):
    """Dump configuraiton file to ``outf``."""
    if not outf:
        outf = sys.stdout

    config_path = pathlib.Path.home() / ".config" / "clinvar-this" / "config.toml"
    if config_path.exists():
        with config_path.open("rt") as configf:
            print(f"# path: {config_path}", file=outf)
            print(configf.read(), file=outf)
    else:
        print(f"# no file at path: {config_path}", file=outf)



"""Support for I/O of the minimal TSV format to define submissions."""

import csv
import datetime
import enum
import pathlib
import re
import typing
import uuid

from logzero import logger
from pydantic import BaseModel
from pydantic.config import ConfigDict

from clinvar_api.models import (
    AffectedStatus,
    AlleleOrigin,
    Assembly,
    Chromosome,
    CitationDb,
    ClinicalSignificanceDescription,
    CollectionMethod,
    ConditionDb,
    ModeOfInheritance,
    MultipleConditionExplanation,
    RecordStatus,
    ReleaseStatus,
    SubmissionAssertionCriteria,
    SubmissionChromosomeCoordinates,
    SubmissionCitation,
    SubmissionClinicalSignificance,
    SubmissionClinvarSubmission,
    SubmissionCondition,
    SubmissionConditionSet,
    SubmissionContainer,
    SubmissionObservedIn,
    SubmissionVariant,
    SubmissionVariantSet,
)
from clinvar_api.models.sub_payload import SubmissionClinicalFeature
from clinvar_api.msg.sub_payload import (
    ClinicalFeaturesAffectedStatus,
    ClinicalFeaturesDb,
    VariantType,
)
from clinvar_this import exceptions


class SeqVarTsvRecord(BaseModel):
    """Record for reading sequence variant TSV."""

    model_config = ConfigDict(frozen=True)

    #: Assembly
    assembly: Assembly
    #: Chromosome
    chromosome: Chromosome
    #: Position
    pos: int
    #: Reference allele bases
    ref: str
    #: Alternative allele bases
    alt: str
    # condition IDs, multiple types supported
    condition: typing.List[str]
    #: Mode of inheritance
    inheritance: typing.Optional[ModeOfInheritance]
    #: Clinical significance
    clinical_significance_description: ClinicalSignificanceDescription
    #: Local identifier of variant-condition pair.
    local_key: typing.Optional[str] = None
    #: Additional columns
    extra_data: typing.Dict[str, str] = {}
    #: Date of last evaluation of clinical significance
    clinical_significance_date_last_evaluated: typing.Optional[str] = None
    #: Additional comment of clinical significance
    clinical_significance_comment: typing.Optional[str] = None
    #: HPO terms for clinical features
    hpo_terms: typing.Optional[typing.List[str]] = None
    #: Pubmed PMID for literature references
    pmids: typing.Optional[typing.List[str]] = None
    #: Existing ClinVar SCV accession
    accession: typing.Optional[str] = None


class StrucVarTsvRecord(BaseModel):
    """Record for reading structural variant TSV."""

    model_config = ConfigDict(frozen=True)

    #: Assembly
    assembly: Assembly
    #: Chromosome
    chromosome: Chromosome
    #: Start position
    start: int
    #: Stop position
    stop: int
    #: Variant type
    sv_type: VariantType
    # condition IDs, multiple types supported
    condition: typing.List[str]
    #: Mode of inheritance
    inheritance: typing.Optional[ModeOfInheritance]
    #: Clinical significance
    clinical_significance_description: ClinicalSignificanceDescription
    #: Local identifier of variant-condition pair.
    local_key: typing.Optional[str] = None
    #: Additional columns
    extra_data: typing.Dict[str, str] = {}
    #: Date of last evaluation of clinical significance
    clinical_significance_date_last_evaluated: typing.Optional[str] = None
    #: Additional comment of clinical significance
    clinical_significance_comment: typing.Optional[str] = None
    #: HPO terms for clinical features
    hpo_terms: typing.Optional[typing.List[str]] = None
    #: Pubmed PMID for literature references
    pmids: typing.Optional[typing.List[str]] = None
    #: Existing ClinVar SCV accession
    accession: typing.Optional[str] = None


class SeqVarHeaderColumn(BaseModel):
    """Header column of sequence variant TSV."""

    model_config = ConfigDict(frozen=True)

    #: Interpreted header names from TSV
    header_names: typing.Tuple[str]
    #: The corresponding key in in ``TsvRecord``
    key: str
    #: Whether the header is required
    required: bool
    #: Type converter on import
    converter: typing.Callable[[str], typing.Any]
    #: Extractor on export
    extractor: typing.Callable[[SeqVarTsvRecord], str]

    @property
    def canonical_name(self):
        """The first entry in ``header_names`` is the canonical one."""
        return self.header_names[0]


class StrucVarHeaderColumn(BaseModel):
    """Header column of structural variant TSV."""

    model_config = ConfigDict(frozen=True)

    #: Interpreted header names from TSV
    header_names: typing.Tuple[str]
    #: The corresponding key in in ``TsvRecord``
    key: str
    #: Whether the header is required
    required: bool
    #: Type converter on import
    converter: typing.Callable[[str], typing.Any]
    #: Extractor on export
    extractor: typing.Callable[[StrucVarTsvRecord], str]

    @property
    def canonical_name(self):
        """The first entry in ``header_names`` is the canonical one."""
        return self.header_names[0]


def _str_list(val: str, pat: str = r"[;,]") -> typing.List[str]:
    """Split a string and return list of trimmed entries"""
    if not val:
        return []
    else:
        return [x.strip() for x in re.split(pat, val)]


def _uuid4_if_falsy(value: typing.Optional[str] = None) -> typing.Union[str, uuid.UUID]:
    """Return a new UUID4 if ``value`` is falsy."""
    if value:
        return value
    else:
        return uuid.uuid4()


def _today_if_falsy(value: typing.Optional[str] = None) -> str:
    """Return string with today's date if ``value`` is falsy."""
    if value:
        return value
    else:
        return datetime.datetime.now().strftime("%Y-%m-%d")


def _enum_value(e: enum.Enum) -> str:
    return str(e.value)


def _enum_value_or_empty(e: typing.Optional[enum.Enum]) -> str:
    if e:
        return str(e.value)
    else:
        return ""


def _join_list(xs: typing.List[typing.Any]) -> str:
    return ",".join([str(x).strip() for x in xs])


#: The header columns for sequence variant TSV files.
SEQ_VAR_HEADER_COLUMNS: typing.Tuple[SeqVarHeaderColumn, ...] = (
    SeqVarHeaderColumn(
        header_names=("ASSEMBLY",),
        key="assembly",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value(r.assembly),
    ),
    SeqVarHeaderColumn(
        header_names=("CHROM",),
        key="chromosome",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value(r.chromosome),
    ),
    SeqVarHeaderColumn(
        header_names=("POS",),
        key="pos",
        required=True,
        converter=int,
        extractor=lambda r: str(r.pos),
    ),
    SeqVarHeaderColumn(
        header_names=("REF",),
        key="ref",
        required=True,
        converter=str,
        extractor=lambda r: str(r.ref),
    ),
    SeqVarHeaderColumn(
        header_names=("ALT",),
        key="alt",
        required=True,
        converter=str,
        extractor=lambda r: str(r.alt),
    ),
    SeqVarHeaderColumn(
        header_names=("CONDITION",),
        key="condition",
        required=True,
        converter=_str_list,
        extractor=lambda r: _join_list(r.condition),
    ),
    SeqVarHeaderColumn(
        header_names=("MOI",),
        key="inheritance",
        required=True,
        converter=lambda x: x or None,
        extractor=lambda r: _enum_value_or_empty(r.inheritance),
    ),
    SeqVarHeaderColumn(
        header_names=("CLIN_SIG",),
        key="clinical_significance_description",
        required=True,
        converter=lambda r: r or None,
        extractor=lambda r: _enum_value_or_empty(r.clinical_significance_description),
    ),
    SeqVarHeaderColumn(
        header_names=("CLIN_EVAL",),
        key="clinical_significance_date_last_evaluated",
        required=False,
        converter=_today_if_falsy,
        extractor=lambda r: str(r.clinical_significance_date_last_evaluated or ""),
    ),
    SeqVarHeaderColumn(
        header_names=("CLIN_COMMENT",),
        key="clinical_significance_comment",
        required=False,
        converter=lambda x: x or None,
        extractor=lambda r: str(r.clinical_significance_comment or ""),
    ),
    SeqVarHeaderColumn(
        header_names=("KEY",),
        key="local_key",
        required=False,
        converter=_uuid4_if_falsy,
        extractor=lambda r: str(r.local_key),
    ),
    SeqVarHeaderColumn(
        header_names=("HPO",),
        key="hpo_terms",
        required=False,
        converter=_str_list,
        extractor=lambda r: _join_list(r.hpo_terms or []),
    ),
    SeqVarHeaderColumn(
        header_names=("PMID",),
        key="pmids",
        required=False,
        converter=_str_list,
        extractor=lambda r: _join_list(r.pmids or []),
    ),
    SeqVarHeaderColumn(
        header_names=("ACCESSION",),
        key="accession",
        required=False,
        converter=str,
        extractor=lambda r: str(r.accession or ""),
    ),
)

#: The header columns for structural variant TSV files.
STRUC_VAR_HEADER_COLUMNS: typing.Tuple[StrucVarHeaderColumn, ...] = (
    StrucVarHeaderColumn(
        header_names=("ASSEMBLY",),
        key="assembly",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value(r.assembly),
    ),
    StrucVarHeaderColumn(
        header_names=("CHROM",),
        key="chromosome",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value(r.chromosome),
    ),
    StrucVarHeaderColumn(
        header_names=("START",),
        key="start",
        required=True,
        converter=int,
        extractor=lambda r: str(r.start),
    ),
    StrucVarHeaderColumn(
        header_names=("STOP",),
        key="stop",
        required=True,
        converter=int,
        extractor=lambda r: str(r.stop),
    ),
    StrucVarHeaderColumn(
        header_names=("SV_TYPE",),
        key="sv_type",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value_or_empty(r.sv_type),
    ),
    StrucVarHeaderColumn(
        header_names=("CONDITION",),
        key="condition",
        required=True,
        converter=_str_list,
        extractor=lambda r: _join_list(r.condition),
    ),
    StrucVarHeaderColumn(
        header_names=("MOI",),
        key="inheritance",
        required=True,
        converter=lambda x: x or None,
        extractor=lambda r: _enum_value_or_empty(r.inheritance),
    ),
    StrucVarHeaderColumn(
        header_names=("CLIN_SIG",),
        key="clinical_significance_description",
        required=True,
        converter=str,
        extractor=lambda r: _enum_value_or_empty(r.clinical_significance_description),
    ),
    StrucVarHeaderColumn(
        header_names=("CLIN_EVAL",),
        key="clinical_significance_date_last_evaluated",
        required=False,
        converter=_today_if_falsy,
        extractor=lambda r: str(r.clinical_significance_date_last_evaluated or ""),
    ),
    StrucVarHeaderColumn(
        header_names=("CLIN_COMMENT",),
        key="clinical_significance_comment",
        required=False,
        converter=lambda x: x or None,
        extractor=lambda r: str(r.clinical_significance_comment or ""),
    ),
    StrucVarHeaderColumn(
        header_names=("KEY",),
        key="local_key",
        required=False,
        converter=_uuid4_if_falsy,
        extractor=lambda r: str(r.local_key),
    ),
    StrucVarHeaderColumn(
        header_names=("HPO",),
        key="hpo_terms",
        required=False,
        converter=_str_list,
        extractor=lambda r: _join_list(r.hpo_terms or []),
    ),
    StrucVarHeaderColumn(
        header_names=("PMID",),
        key="hpo_terms",
        required=False,
        converter=_str_list,
        extractor=lambda r: _join_list(r.pmids or []),
    ),
    StrucVarHeaderColumn(
        header_names=("ACCESSION",),
        key="accession",
        required=False,
        converter=str,
        extractor=lambda r: str(r.accession or ""),
    ),
)


class TsvType(enum.Enum):
    """Type of TSV file."""

    #: Sequence variants.
    SEQ_VAR = "seqvar"
    #: Structural variants.
    STRUC_VAR = "strucvar"


def guess_tsv_type(path: str) -> typing.Optional[TsvType]:
    """Guess TSV type."""
    with open(path, "rt") as inputf:
        arr = inputf.readline().strip().split("\t")
    try:
        _map_seq_var_header(arr)
        return TsvType.SEQ_VAR
    except exceptions.InvalidFormat:
        try:
            _map_struc_var_header(arr)
            return TsvType.STRUC_VAR
        except exceptions.InvalidFormat:
            return None


def _map_seq_var_header(
    header: typing.List[str],
) -> typing.List[typing.Optional[SeqVarHeaderColumn]]:
    """Map header row from sequence variant TSV file to header columns

    Map to ``None`` for extra data columns.  Raises if a required column is missing.
    """
    seen_required = {
        column.canonical_name: False for column in SEQ_VAR_HEADER_COLUMNS if column.required
    }
    by_name = {name: column for column in SEQ_VAR_HEADER_COLUMNS for name in column.header_names}
    result = []
    for entry in header:
        column = by_name.get(entry)
        if column:
            seen_required[column.canonical_name] = True
        result.append(column)

    missing_columns = [name for name, seen in seen_required.items() if not seen]
    if missing_columns:
        raise exceptions.InvalidFormat(f"Missing columns in TSV file: {missing_columns}")

    return result


def _map_struc_var_header(
    header: typing.List[str],
) -> typing.List[typing.Optional[StrucVarHeaderColumn]]:
    """Map header row from structural variant TSV file to header columns

    Map to ``None`` for extra data columns.  Raises if a required column is missing.
    """
    seen_required = {
        column.canonical_name: False for column in STRUC_VAR_HEADER_COLUMNS if column.required
    }
    by_name = {name: column for column in STRUC_VAR_HEADER_COLUMNS for name in column.header_names}
    result = []
    for entry in header:
        column = by_name.get(entry)
        if column:
            seen_required[column.canonical_name] = True
        result.append(column)

    missing_columns = [name for name, seen in seen_required.items() if not seen]
    if missing_columns:
        raise exceptions.InvalidFormat(f"Missing columns in TSV file: {missing_columns}")

    return result


def _read_seq_var_tsv_file(inputf: typing.TextIO) -> typing.List[SeqVarTsvRecord]:
    """Read sequence variant TSV from the given file."""

    def row_empty(row: typing.List[str]) -> bool:
        return not row or not [val.strip() for val in row if val.strip()]

    reader = csv.reader(inputf, delimiter="\t")
    header_row = None
    headers = None

    result: typing.List[SeqVarTsvRecord] = []
    for lineno, row in enumerate(reader):
        if row_empty(row):
            continue  # skip empty lines
        if header_row:
            raw_record = {}
            extra_data = {}
            if len(row) != len(header_row):
                raise exceptions.InvalidFormat(f"Wrong number of rows in line {lineno+1}")
            for value, header, header_name in zip(row, headers, header_row):
                if header:
                    raw_record[header.key] = header.converter(value)
                else:
                    extra_data[header_name] = value
            record = SeqVarTsvRecord.model_validate(raw_record)
            result.append(record.model_copy(update={"extra_data": extra_data}))
        else:
            header_row = row
            headers = _map_seq_var_header(row)
    return result


def _read_struc_var_tsv_file(inputf: typing.TextIO) -> typing.List[StrucVarTsvRecord]:
    """Read structural variant TSV from the given file."""

    def row_empty(row: typing.List[str]) -> bool:
        return not row or not [val.strip() for val in row if val.strip()]

    reader = csv.reader(inputf, delimiter="\t")
    header_row = None
    headers = None

    result: typing.List[StrucVarTsvRecord] = []
    for lineno, row in enumerate(reader):
        if row_empty(row):
            continue  # skip empty lines
        if header_row:
            raw_record = {}
            extra_data = {}
            if len(row) != len(header_row):
                raise exceptions.InvalidFormat(f"Wrong number of rows in line {lineno+1}")
            for value, header, header_name in zip(row, headers, header_row):
                if header:
                    raw_record[header.key] = header.converter(value)
                else:
                    extra_data[header_name] = value
            record = StrucVarTsvRecord.model_validate(raw_record)
            result.append(record.model_copy(update={"extra_data": extra_data}))
        else:
            header_row = row
            headers = _map_struc_var_header(row)
    return result


def read_seq_var_tsv(
    *,
    file: typing.Optional[typing.TextIO] = None,
    path: typing.Union[None, str, pathlib.Path] = None,
) -> typing.List[SeqVarTsvRecord]:
    """Read sequence variant TSV from either file or path"""
    if file:
        return _read_seq_var_tsv_file(file)
    elif path:
        with pathlib.Path(path).open("rt") as inputf:
            return _read_seq_var_tsv_file(inputf)
    else:
        raise TypeError("You have to provide either file or path")


def read_struc_var_tsv(
    *,
    file: typing.Optional[typing.TextIO] = None,
    path: typing.Union[None, str, pathlib.Path] = None,
) -> typing.List[StrucVarTsvRecord]:
    """Read structural variant TSV from either file or path"""
    if file:
        return _read_struc_var_tsv_file(file)
    elif path:
        with pathlib.Path(path).open("rt") as inputf:
            return _read_struc_var_tsv_file(inputf)
    else:
        raise TypeError("You have to provide either file or path")


def _write_seq_var_tsv_file(tsv_records: typing.Iterable[SeqVarTsvRecord], outputf: typing.TextIO):
    """Write sequence variant records as TSV to the given file."""
    extra_keys = []
    for record in tsv_records:
        if record.extra_data:
            for key in record.extra_data:
                if key not in extra_keys:
                    extra_keys.append(key)
    writer = csv.writer(outputf, delimiter="\t")
    writer.writerow([h.canonical_name for h in SEQ_VAR_HEADER_COLUMNS] + extra_keys)
    for record in tsv_records:
        row = [hc.extractor(record) for hc in SEQ_VAR_HEADER_COLUMNS] + [
            record.extra_data.get(extra_key, "") for extra_key in extra_keys
        ]
        writer.writerow(row)


def _write_struc_var_tsv_file(
    tsv_records: typing.Iterable[StrucVarTsvRecord], outputf: typing.TextIO
):
    """Write structural variant records as TSV to the given file."""
    extra_keys = []
    for record in tsv_records:
        if record.extra_data:
            for key in record.extra_data:
                if key not in extra_keys:
                    extra_keys.append(key)
    writer = csv.writer(outputf, delimiter="\t")
    writer.writerow([h.canonical_name for h in STRUC_VAR_HEADER_COLUMNS] + extra_keys)
    for record in tsv_records:
        row = [hc.extractor(record) for hc in STRUC_VAR_HEADER_COLUMNS] + [
            record.extra_data.get(extra_key, "") for extra_key in extra_keys
        ]
        writer.writerow(row)


def write_seq_var_tsv(
    tsv_records: typing.Iterable[SeqVarTsvRecord],
    *,
    file: typing.Optional[typing.TextIO] = None,
    path: typing.Union[None, str, pathlib.Path] = None,
):
    """Write sequence variant TSV to either file or path"""
    if file:
        return _write_seq_var_tsv_file(tsv_records, file)
    elif path:
        with pathlib.Path(path).open("wt") as outputf:
            return _write_seq_var_tsv_file(tsv_records, outputf)
    else:
        raise TypeError("You have to provide either file or path")


def write_struc_var_tsv(
    tsv_records: typing.Iterable[StrucVarTsvRecord],
    *,
    file: typing.Optional[typing.TextIO] = None,
    path: typing.Union[None, str, pathlib.Path] = None,
):
    """Write structural variant TSV to either file or path"""
    if file:
        return _write_struc_var_tsv_file(tsv_records, file)
    elif path:
        with pathlib.Path(path).open("wt") as outputf:
            return _write_struc_var_tsv_file(tsv_records, outputf)
    else:
        raise TypeError("You have to provide either file or path")


class BatchMetadata(BaseModel):
    """Batch-wide settings for TSV import.

    The properties will be assigned to all variants/samples in the batch.
    """

    model_config = ConfigDict(frozen=True)

    #: The collection method
    collection_method: typing.Optional[CollectionMethod] = None
    #: The allele origin
    allele_origin: typing.Optional[AlleleOrigin] = None
    #: The release status
    release_status: typing.Optional[ReleaseStatus] = None


#: Default values for ``BatchMetadata`` to use optional.
BATCH_METADATA_DEFAULTS: typing.Dict[str, typing.Any] = {
    "collection_method": CollectionMethod.NOT_PROVIDED,
    "allele_origin": AlleleOrigin.GERMLINE,
    "release_status": ReleaseStatus.PUBLIC,
}


def batch_metadata_from_mapping(
    keys_values: typing.Iterable[str], *, use_defaults: bool = False
) -> BatchMetadata:
    """Convert configuration from ``KEY=VALUE`` strings to ``BatchMetadata``

    Default values can be used (should be on import but not on update).
    """
    field_types = {name: value for (name, value) in typing.get_type_hints(BatchMetadata).items()}
    kwargs = {}
    for key_value in keys_values:
        if "=" not in key_value:
            raise exceptions.ArgumentsError(f"Invalid key/value pair in {key_value}")
        key, value = key_value.split("=")
        if key in field_types:
            try:
                # We need to ignore types as mypy 1.6.0 warns for "expected type[any] but
                # found "type[any] | None".
                kwargs[key] = field_types[key].model_validate(value)  # type: ignore
            except ValueError:
                raise exceptions.ArgumentsError(f"Failed to parse {value} as for key {key}")
    if use_defaults:
        for key, value in BATCH_METADATA_DEFAULTS.items():
            kwargs.setdefault(key, value)
    return BatchMetadata(**kwargs)


def select_condition_db(condition_id):
    if "OMIM" in condition_id:
        return ConditionDb.OMIM
    elif "ORPHA" in condition_id:
        return ConditionDb.ORPHANET
    elif "MONDO" in condition_id:
        return ConditionDb.MONDO
    elif "HP" in condition_id:
        return ConditionDb.HP
    else:
        raise TypeError(
            f"Failed to match condition to supported condition types: OMIM, ORPHA, MONDO for {condition_id}"
        )


def get_condition_id_fmt(condition_id: str) -> str:
    condition_db = select_condition_db(condition_id)
    if condition_db == ConditionDb.OMIM:
        return condition_id.split(":")[1].strip()
    elif condition_db == ConditionDb.MONDO:
        return condition_id
    elif condition_db == ConditionDb.ORPHANET:
        return "ORPHA" + condition_id.split(":")[1].strip()
    elif condition_db == ConditionDb.HP:
        return condition_id
    else:
        raise TypeError(
            f"Failed to match condition to supported condition types: OMIM, ORPHA, MONDO for {condition_id}"
        )


def condition_to_fmt_string(condition: SubmissionCondition) -> str:
    condition_id = condition.id or ""
    if condition.db == ConditionDb.OMIM:
        return f"OMIM:{condition.id}"
    elif condition.db == ConditionDb.MONDO:
        return condition_id
    elif condition.db == ConditionDb.HP:
        return condition_id
    elif condition.db == ConditionDb.ORPHANET:
        return f"ORPHA:{condition_id.replace('ORPHA', '')}"
    elif condition.name == "not provided":
        return ""
    else:
        raise TypeError(
            f"Failed to match condition to supported condition types: OMIM, ORPHA, MONDO for {condition.db}"
        )


def record_conditions(
    record: typing.Union[SeqVarTsvRecord, StrucVarTsvRecord]
) -> typing.List[SubmissionCondition]:
    """Construct ``SubmissionCondition`` from ``TsvRecord``."""
    if not record.condition or record.condition == ["not provided"]:
        return [SubmissionCondition(name="not provided")]
    else:
        explanation_values = [c.value for c in MultipleConditionExplanation]
        return [
            SubmissionCondition(
                db=select_condition_db(condition_id), id=get_condition_id_fmt(condition_id)
            )
            for condition_id in record.condition
            if condition_id not in explanation_values
        ]


def record_condition_explanation(
    record: typing.Union[SeqVarTsvRecord, StrucVarTsvRecord],
) -> typing.Optional[MultipleConditionExplanation]:
    explanation_values = [c.value for c in MultipleConditionExplanation]
    for condition_id in record.condition:
        if condition_id in explanation_values:
            return MultipleConditionExplanation(condition_id)
    return None


def record_condition_set(
    record: typing.Union[SeqVarTsvRecord, StrucVarTsvRecord]
) -> SubmissionConditionSet:
    conditions = record_conditions(record)
    multiple_condition_explanation = record_condition_explanation(record)
    return SubmissionConditionSet(
        condition=conditions, multiple_condition_explanation=multiple_condition_explanation
    )


def seq_var_tsv_records_to_submission_container(
    tsv_records: typing.List[SeqVarTsvRecord],
    batch_metadata: BatchMetadata,
) -> SubmissionContainer:
    """Convert seq. var. TSV records to submission container data structure."""

    def record_pubmed_citations(
        record: SeqVarTsvRecord,
    ) -> typing.Optional[typing.List[SubmissionCitation]]:
        if record.pmids:
            return [
                SubmissionCitation(
                    db=CitationDb.PUBMED,
                    id=pmid,
                )
                for pmid in record.pmids
            ]
        else:
            return None

    def record_clinical_features(
        record: SeqVarTsvRecord,
    ) -> typing.Optional[typing.List[SubmissionClinicalFeature]]:
        """Construct ``typing.Optional[typing.List[SubmissionClinicalFeature]]`` from ``TsvRecord``."""
        if record.hpo_terms:
            return [
                SubmissionClinicalFeature(
                    clinical_features_affected_status=ClinicalFeaturesAffectedStatus.PRESENT,
                    db=ClinicalFeaturesDb.HP,
                    id=hpo_term,
                )
                for hpo_term in record.hpo_terms
            ]
        else:
            return None

    allele_origin = batch_metadata.allele_origin or BATCH_METADATA_DEFAULTS["allele_origin"]
    collection_method = (
        batch_metadata.collection_method or BATCH_METADATA_DEFAULTS["collection_method"]
    )
    release_status = batch_metadata.release_status or BATCH_METADATA_DEFAULTS["release_status"]

    return SubmissionContainer(
        assertion_criteria=SubmissionAssertionCriteria(
            # The following should come from the profile, cf.
            #
            # https://github.com/bihealth/clinvar-this/issues/36
            db=CitationDb.PUBMED,
            id="25741868",
        ),
        clinvar_submission_release_status=release_status,
        clinvar_submission=[
            SubmissionClinvarSubmission(
                clinvar_accession=record.accession,
                local_id=str(_uuid4_if_falsy()),
                local_key=record.local_key,
                condition_set=record_condition_set(record),
                observed_in=[
                    SubmissionObservedIn(
                        affected_status=AffectedStatus.YES,
                        allele_origin=allele_origin,
                        collection_method=collection_method,
                        clinical_features=record_clinical_features(record),
                    )
                ],
                clinical_significance=SubmissionClinicalSignificance(
                    clinical_significance_description=record.clinical_significance_description,
                    mode_of_inheritance=record.inheritance,
                    citation=record_pubmed_citations(record),
                ),
                record_status=RecordStatus.NOVEL,
                variant_set=SubmissionVariantSet(
                    variant=[
                        SubmissionVariant(
                            chromosome_coordinates=SubmissionChromosomeCoordinates(
                                assembly=record.assembly,
                                chromosome=record.chromosome,
                                start=record.pos,
                                stop=record.pos + len(record.ref) - 1,
                                reference_allele=record.ref,
                                alternate_allele=record.alt,
                            ),
                        )
                    ]
                ),
                extra_data=record.extra_data or None,  # prefer ``None`` over ``{}``
            )
            for record in tsv_records
        ],
    )


def struc_var_tsv_records_to_submission_container(
    tsv_records: typing.List[StrucVarTsvRecord],
    batch_metadata: BatchMetadata,
) -> SubmissionContainer:
    """Convert struc. var. TSV records to submission container data structure."""

    def record_pubmed_citations(
        record: StrucVarTsvRecord,
    ) -> typing.Optional[typing.List[SubmissionCitation]]:
        if record.pmids:
            return [
                SubmissionCitation(
                    db=CitationDb.PUBMED,
                    id=pmid,
                )
                for pmid in record.pmids
            ]
        else:
            return None

    def record_clinical_features(
        record: StrucVarTsvRecord,
    ) -> typing.Optional[typing.List[SubmissionClinicalFeature]]:
        """Construct ``typing.Optional[typing.List[SubmissionClinicalFeature]]`` from ``TsvRecord``."""
        if record.hpo_terms:
            return [
                SubmissionClinicalFeature(
                    clinical_features_affected_status=ClinicalFeaturesAffectedStatus.PRESENT,
                    db=ClinicalFeaturesDb.HP,
                    id=hpo_term,
                )
                for hpo_term in record.hpo_terms
            ]
        else:
            return None

    allele_origin = batch_metadata.allele_origin or BATCH_METADATA_DEFAULTS["allele_origin"]
    collection_method = (
        batch_metadata.collection_method or BATCH_METADATA_DEFAULTS["collection_method"]
    )
    release_status = batch_metadata.release_status or BATCH_METADATA_DEFAULTS["release_status"]

    return SubmissionContainer(
        assertion_criteria=SubmissionAssertionCriteria(
            # The following should come from the profile, cf.
            #
            # https://github.com/bihealth/clinvar-this/issues/36
            db=CitationDb.PUBMED,
            id="25741868",
        ),
        clinvar_submission_release_status=release_status,
        clinvar_submission=[
            SubmissionClinvarSubmission(
                clinvar_accession=record.accession,
                local_id=str(_uuid4_if_falsy()),
                local_key=record.local_key,
                condition_set=record_condition_set(record),
                observed_in=[
                    SubmissionObservedIn(
                        affected_status=AffectedStatus.YES,
                        allele_origin=allele_origin,
                        collection_method=collection_method,
                        clinical_features=record_clinical_features(record),
                    )
                ],
                clinical_significance=SubmissionClinicalSignificance(
                    clinical_significance_description=record.clinical_significance_description,
                    mode_of_inheritance=record.inheritance,
                    citation=record_pubmed_citations(record),
                ),
                record_status=RecordStatus.NOVEL,
                variant_set=SubmissionVariantSet(
                    variant=[
                        SubmissionVariant(
                            chromosome_coordinates=SubmissionChromosomeCoordinates(
                                assembly=record.assembly,
                                chromosome=record.chromosome,
                                start=record.start,
                                stop=record.stop,
                            ),
                            variant_type=record.sv_type,
                        )
                    ]
                ),
                extra_data=record.extra_data or None,  # prefer ``None`` over ``{}``
            )
            for record in tsv_records
        ],
    )


def format_conditions(condition_set: SubmissionConditionSet) -> typing.List[str]:
    conditions = condition_set.condition
    fmt_conditions = []
    if conditions:
        for condition in conditions:
            if c := condition_to_fmt_string(condition):
                fmt_conditions.append(c)
    if condition_set.multiple_condition_explanation:
        fmt_conditions.append(condition_set.multiple_condition_explanation.value)
    return fmt_conditions


def format_hpo_terms(submission: SubmissionClinvarSubmission) -> typing.Optional[typing.List[str]]:
    clinical_features = submission.observed_in[0].clinical_features
    result = None
    if clinical_features:
        result = [hpo_term.id for hpo_term in clinical_features if hpo_term.id]
    return result


def submission_container_to_seq_var_tsv_records(  # noqa: C901
    submission_container: SubmissionContainer,
) -> typing.List[SeqVarTsvRecord]:
    def _inheritance(submission: SubmissionClinvarSubmission) -> typing.Optional[ModeOfInheritance]:
        if submission.clinical_significance.mode_of_inheritance:
            return submission.clinical_significance.mode_of_inheritance
        else:
            return None

    def _pmids(submission: SubmissionClinvarSubmission) -> typing.Optional[typing.List[str]]:
        if citations := submission.clinical_significance.citation:
            return [c.id for c in citations if c.db == CitationDb.PUBMED and c.id]
        else:
            return None

    def submission_to_seq_var_tsv_record(
        submission: SubmissionClinvarSubmission,
    ) -> typing.Optional[SeqVarTsvRecord]:
        if not submission.variant_set:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no variant set"
            )
        elif not submission.variant_set.variant:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no variant"
            )
        elif not submission.variant_set.variant[0].chromosome_coordinates:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no chromosome coordinates"
            )
        else:
            chromosome_coordinates: SubmissionChromosomeCoordinates = (
                submission.variant_set.variant[0].chromosome_coordinates
            )
        if not (chromosome_coordinates.assembly and chromosome_coordinates.chromosome):
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - missing chromosome"
            )
        if submission.variant_set.variant[0].variant_type:
            logger.warning(
                "Skipping variant at %s:%d as it looks like a SV",
                chromosome_coordinates.chromosome.value,
                chromosome_coordinates.start,
            )
            return None
        if not (
            chromosome_coordinates.start
            and chromosome_coordinates.reference_allele
            and chromosome_coordinates.alternate_allele
        ):
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - incomplete coordinates"
            )

        extra_data = {}
        if submission.extra_data:
            extra_data.update(submission.extra_data)

        return SeqVarTsvRecord(
            accession=submission.clinvar_accession,
            assembly=chromosome_coordinates.assembly,
            chromosome=chromosome_coordinates.chromosome,
            pos=chromosome_coordinates.start,
            ref=chromosome_coordinates.reference_allele,
            alt=chromosome_coordinates.alternate_allele,
            condition=format_conditions(submission.condition_set),
            inheritance=_inheritance(submission),
            clinical_significance_description=submission.clinical_significance.clinical_significance_description,
            pmids=_pmids(submission),
            local_key=submission.local_key or "",
            extra_data=extra_data,
            clinical_significance_date_last_evaluated=submission.clinical_significance.date_last_evaluated
            or "",
            clinical_significance_comment=submission.clinical_significance.comment or "",
            hpo_terms=format_hpo_terms(submission),
        )

    clinvar_submissions = submission_container.clinvar_submission or []

    result: typing.List[SeqVarTsvRecord] = []
    for submission in clinvar_submissions:
        record = submission_to_seq_var_tsv_record(submission)
        if record:
            result.append(record)
    return result


def submission_container_to_struc_var_tsv_records(  # noqa: C901
    submission_container: SubmissionContainer,
) -> typing.List[StrucVarTsvRecord]:
    def _inheritance(submission: SubmissionClinvarSubmission) -> typing.Optional[ModeOfInheritance]:
        if submission.clinical_significance.mode_of_inheritance:
            return submission.clinical_significance.mode_of_inheritance
        else:
            return None

    def _pmids(submission: SubmissionClinvarSubmission) -> typing.Optional[typing.List[str]]:
        if citations := submission.clinical_significance.citation:
            return [c.id for c in citations if c.db == CitationDb.PUBMED and c.id]
        else:
            return None

    def submission_to_struc_var_tsv_record(
        submission: SubmissionClinvarSubmission,
    ) -> typing.Optional[StrucVarTsvRecord]:
        if not submission.variant_set:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no variant set"
            )
        elif not submission.variant_set.variant:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no variant"
            )
        elif not submission.variant_set.variant[0].chromosome_coordinates:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no chromosome coordinates"
            )
        else:
            chromosome_coordinates: SubmissionChromosomeCoordinates = (
                submission.variant_set.variant[0].chromosome_coordinates
            )
            variant_type: typing.Optional[VariantType] = submission.variant_set.variant[
                0
            ].variant_type
        if not (chromosome_coordinates.assembly and chromosome_coordinates.chromosome):
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - missing chromosome"
            )
        if not submission.variant_set.variant[0].variant_type:
            logger.warning(
                "Skipping variant at %s:%d as it does not look like a SV",
                chromosome_coordinates.chromosome.value,
                chromosome_coordinates.start,
            )
            return None
        if not (chromosome_coordinates.start and chromosome_coordinates.stop):
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - incomplete coordinates"
            )
        if not variant_type:
            raise exceptions.ClinvarThisException(
                "Problem with internal data structure - no variant type"
            )

        extra_data = {}
        if submission.extra_data:
            extra_data.update(submission.extra_data)

        return StrucVarTsvRecord(
            accession=submission.clinvar_accession,
            assembly=chromosome_coordinates.assembly,
            chromosome=chromosome_coordinates.chromosome,
            start=chromosome_coordinates.start,
            stop=chromosome_coordinates.stop,
            sv_type=variant_type,
            condition=format_conditions(submission.condition_set),
            pmids=_pmids(submission),
            inheritance=_inheritance(submission),
            clinical_significance_description=submission.clinical_significance.clinical_significance_description,
            local_key=submission.local_key or "",
            extra_data=extra_data,
            clinical_significance_date_last_evaluated=submission.clinical_significance.date_last_evaluated
            or "",
            clinical_significance_comment=submission.clinical_significance.comment or "",
            hpo_terms=format_hpo_terms(submission),
        )

    clinvar_submissions = submission_container.clinvar_submission or []

    result: typing.List[StrucVarTsvRecord] = []
    for submission in clinvar_submissions:
        record = submission_to_struc_var_tsv_record(submission)
        if record:
            result.append(record)
    return result



"""Determin ACMG class by frequency."""

import enum
import gzip
import json
import typing

import tqdm

from clinvar_data import models

DEFAULT_THRESHOLDS = [
    0.0,
    1e-05,
    2.5e-05,
    5e-05,
    0.0001,
    0.00025,
    0.0005,
    0.001,
    0.0025,
    0.005,
    0.01,
    0.025,
    0.05,
    0.1,
    0.25,
    0.5,
    1.0,
]


@enum.unique
class Classes(enum.Enum):
    BENIGN = "benign"
    UNCERTAIN = "uncertain"
    PATHOGENIC = "pathogenic"


CLINSIG_TO_CLASS = {
    models.ClinicalSignificanceDescription.BENIGN: Classes.BENIGN,
    models.ClinicalSignificanceDescription.LIKELY_BENIGN: Classes.BENIGN,
    models.ClinicalSignificanceDescription.UNCERTAIN_SIGNIFICANCE: Classes.UNCERTAIN,
    models.ClinicalSignificanceDescription.LIKELY_PATHOGENIC: Classes.PATHOGENIC,
    models.ClinicalSignificanceDescription.PATHOGENIC: Classes.PATHOGENIC,
}


def zero_counts(count: int):
    return {klass: [0] * (count + 1) for klass in Classes}


def locate(thresholds: typing.List[float], value: typing.Optional[float]) -> int:
    if value is None:
        return 0
    for i, threshold in enumerate(thresholds):
        if value <= threshold:
            return i
    return len(thresholds)


def generate_counts(path_input: str, thresholds: typing.List[float]):
    counts = {}

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    with inputf:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            clinvar_set = models.ClinVarSet.model_validate_json(line)

            pathogenicity = (
                clinvar_set.reference_clinvar_assertion.clinical_significance.description
            )

            if not clinvar_set.reference_clinvar_assertion.measures:
                continue
            freq = None
            for measure in clinvar_set.reference_clinvar_assertion.measures.measures:
                if measure.global_minor_allele_frequency:
                    freq = measure.global_minor_allele_frequency.value

            hgnc = None
            for measure in clinvar_set.reference_clinvar_assertion.measures.measures:
                for measure_relationship in measure.measure_relationship:
                    for xref in measure_relationship.xrefs:
                        if xref.db == "HGNC":
                            hgnc = xref.id
                            break

            if not hgnc or pathogenicity not in CLINSIG_TO_CLASS:
                continue

            if hgnc not in counts:
                counts[hgnc] = zero_counts(len(thresholds))

            idx = locate(thresholds, freq)
            counts[hgnc][CLINSIG_TO_CLASS[pathogenicity]][idx] += 1

    return counts


def write_report(counts: dict, path_output: str):
    if path_output.endswith(".gz"):
        outputf = gzip.open(path_output, "wt")
    else:
        outputf = open(path_output, "wt")

    with outputf:
        for hgnc, counts in sorted(counts.items(), key=lambda x: int(x[0][5:])):
            print(
                json.dumps({"hgnc": hgnc, "counts": counts.model_dump(mode="json")}),  # type: ignore
                file=outputf,
            )


def run_report(path_input: str, path_output: str, thresholds: typing.List[float]):
    counts = generate_counts(path_input, thresholds)
    write_report(counts, path_output)



"""Helpers for schema validation."""

import json
import pathlib
import typing

from jsonschema import validate


def validate_submission_payload(payload: typing.Any):
    schema_path = pathlib.Path(__file__).parent / "submission_schema.json"
    with schema_path.open("rt") as inputf:
        schema = json.load(inputf)

    validate(instance=payload, schema=schema)


def validate_status_summary(payload: typing.Any):
    schema_path = pathlib.Path(__file__).parent / "summary_response_schema.json"
    with schema_path.open("rt") as inputf:
        schema = json.load(inputf)

    validate(instance=payload, schema=schema)



"""Data structures for parsing submission response."""

from enum import Enum
import typing

from pydantic import BaseModel
from pydantic.config import ConfigDict


class BatchProcessingStatus(Enum):
    IN_PROCESSING = "In processing"
    SUCCESS = "Success"
    ERROR = "Error"
    PARTIAL_SUCCESS = "Partial success"


class BatchReleaseStatus(Enum):
    RELEASED = "Released"
    PARTIAL_RELEASED = "Partial released"
    NOT_RELEASED = "Not released"


class Created(BaseModel):
    """Representation of successful creation."""

    model_config = ConfigDict(frozen=True)

    #: The submission ID.
    id: str


class Error(BaseModel):
    """Representation of server's response in case of failure."""

    model_config = ConfigDict(frozen=True)

    #: The error response's message.
    message: str


class SummaryResponseErrorInput(BaseModel):
    model_config = ConfigDict(frozen=True)

    value: typing.Optional[str] = None
    field: typing.Optional[str] = None


class SummaryResponseErrorOutputError(BaseModel):
    model_config = ConfigDict(frozen=True)

    userMessage: str


class SummaryResponseErrorOutput(BaseModel):
    model_config = ConfigDict(frozen=True)

    errors: typing.List[SummaryResponseErrorOutputError]


class SummaryResponseError(BaseModel):
    model_config = ConfigDict(frozen=True)

    # NB: docs and schema say required but examples do not show
    input: typing.List[SummaryResponseErrorInput]
    output: SummaryResponseErrorOutput


class SummaryResponseDeletionIdentifier(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinvarAccession: str
    clinvarLocalKey: typing.Optional[str] = None


class SummaryResponseDeletion(BaseModel):
    model_config = ConfigDict(frozen=True)

    identifiers: SummaryResponseDeletionIdentifier
    processingStatus: str
    deleteDate: typing.Optional[str] = None
    deleteStatus: typing.Optional[str] = None
    errors: typing.Optional[typing.List[SummaryResponseError]] = None


class SummaryResponseSubmissionIdentifiers(BaseModel):
    model_config = ConfigDict(frozen=True)

    clinvarLocalKey: str
    clinvarAccession: typing.Optional[str] = None
    localID: typing.Optional[str] = None
    localKey: typing.Optional[str] = None


class SummaryResponseSubmission(BaseModel):
    model_config = ConfigDict(frozen=True)

    identifiers: SummaryResponseSubmissionIdentifiers
    processingStatus: str
    clinvarAccessionVersion: typing.Optional[str] = None
    errors: typing.Optional[typing.List[SummaryResponseError]] = None
    releaseDate: typing.Optional[str] = None
    releaseStatus: typing.Optional[str] = None


class SummaryResponse(BaseModel):
    """Represetation of server's response to a submission."""

    model_config = ConfigDict(frozen=True)

    batchProcessingStatus: BatchProcessingStatus
    batchReleaseStatus: BatchReleaseStatus
    submissionDate: str
    submissionName: str
    totalCount: int
    totalErrors: int
    totalPublic: int
    totalSuccess: int
    deletions: typing.Optional[typing.List[SummaryResponseDeletion]] = None
    submissions: typing.Optional[typing.List[SummaryResponseSubmission]] = None
    totalDeleteCount: typing.Optional[int] = None
    totalDeleted: typing.Optional[int] = None
    totalDeleteErrors: typing.Optional[int] = None
    totalDeleteSuccess: typing.Optional[int] = None
